@@ -63,12 +63,12 @@
 
 
 Datum geom_from_gml(PG_FUNCTION_ARGS);
-static LWGEOM* lwgeom_from_gml(const char *wkt);
+static LWGEOM *lwgeom_from_gml(const char *wkt, int xml_size);
 static LWGEOM* parse_gml(xmlNodePtr xnode, bool *hasz, int *root_srid);
 
 typedef struct struct_gmlSrs
 {
-	int srid;
+	int32_t srid;
 	bool reverse_axis;
 }
 gmlSrs;
@@ -102,23 +102,18 @@ Datum geom_from_gml(PG_FUNCTION_ARGS)
 	LWGEOM *lwgeom;
 	char *xml;
 	int root_srid=SRID_UNKNOWN;
-
+	int xml_size;
 
 	/* Get the GML stream */
 	if (PG_ARGISNULL(0)) PG_RETURN_NULL();
 	xml_input = PG_GETARG_TEXT_P(0);
 	xml = text_to_cstring(xml_input);
+	xml_size = VARSIZE_ANY_EXHDR(xml_input);
 
 	/* Zero for undefined */
 	root_srid = PG_GETARG_INT32(1);
 
-	/* Internally lwgeom_from_gml calls gml_reproject_pa which, for PROJ before 6, called GetProj4String.
-	 * That function requires access to spatial_ref_sys, so in order to have it ready we need to ensure
-	 * the internal cache is initialized
-	 */
-	postgis_initialize_cache(fcinfo);
-	lwgeom = lwgeom_from_gml(xml);
-
+	lwgeom = lwgeom_from_gml(xml, xml_size);
 	if ( root_srid != SRID_UNKNOWN )
 		lwgeom->srid = root_srid;
 
@@ -297,43 +292,96 @@ static xmlNodePtr get_xlink_node(xmlNodePtr xnode)
 }
 
 
+
 /**
- * Use Proj4 to reproject a given POINTARRAY
+ * Use Proj to reproject a given POINTARRAY
  */
-static POINTARRAY* gml_reproject_pa(POINTARRAY *pa, int srid_in, int srid_out)
+
+#if POSTGIS_PROJ_VERSION < 60
+
+static POINTARRAY *
+gml_reproject_pa(POINTARRAY *pa, int32_t srid_in, int32_t srid_out)
 {
-	projPJ in_pj, out_pj;
+	PJ pj;
 	char *text_in, *text_out;
 
 	if (srid_in == SRID_UNKNOWN) return pa; /* nothing to do */
 	if (srid_out == SRID_UNKNOWN) gml_lwpgerror("invalid GML representation", 3);
 
-	text_in = GetProj4StringSPI(srid_in);
-	text_out = GetProj4StringSPI(srid_out);
+	text_in = GetProj4String(srid_in);
+	text_out = GetProj4String(srid_out);
 
-	in_pj = lwproj_from_string(text_in);
-	out_pj = lwproj_from_string(text_out);
+	pj.pj_from = projpj_from_string(text_in);
+	pj.pj_to = projpj_from_string(text_out);
 
 	lwfree(text_in);
 	lwfree(text_out);
 
-	if ( ptarray_transform(pa, in_pj, out_pj) == LW_FAILURE )
+	if ( ptarray_transform(pa, &pj) == LW_FAILURE )
 	{
 		elog(ERROR, "gml_reproject_pa: reprojection failed");
 	}
 
-	pj_free(in_pj);
-	pj_free(out_pj);
+	pj_free(pj.pj_from);
+	pj_free(pj.pj_to);
 
 	return pa;
 }
+#else
+/*
+ * TODO: rework GML projection handling to skip the spatial_ref_sys
+ * lookups, and use the Proj 6+ EPSG catalogue and built-in SRID
+ * lookups directly. Drop this ugly hack.
+ */
+static POINTARRAY *
+gml_reproject_pa(POINTARRAY *pa, int32_t epsg_in, int32_t epsg_out)
+{
+	PJ *pj;
+	LWPROJ *lwp;
+	char text_in[16];
+	char text_out[16];
+
+	if (epsg_in == SRID_UNKNOWN)
+		return pa; /* nothing to do */
+
+	if (epsg_out == SRID_UNKNOWN)
+	{
+		gml_lwpgerror("invalid GML representation", 3);
+		return NULL;
+	}
+
+	snprintf(text_in, 16, "EPSG:%d", epsg_in);
+	snprintf(text_out, 16, "EPSG:%d", epsg_out);
+	pj = proj_create_crs_to_crs(NULL, text_in, text_out, NULL);
+
+	lwp = lwproj_from_PJ(pj, LW_FALSE);
+	if (!lwp)
+	{
+		proj_destroy(pj);
+		gml_lwpgerror("Could not create LWPROJ*", 57);
+		return NULL;
+	}
+
+	if (ptarray_transform(pa, lwp) == LW_FAILURE)
+	{
+		proj_destroy(pj);
+		elog(ERROR, "gml_reproject_pa: reprojection failed");
+		return NULL;
+	}
+	proj_destroy(pj);
+	pfree(lwp);
+
+	return pa;
+}
+#endif /* POSTGIS_PROJ_VERSION */
 
 
 /**
  * Return 1 if given srid is planar (0 otherwise, i.e geocentric srid)
  * Return -1 if srid is not in spatial_ref_sys
  */
-static int gml_is_srid_planar(int srid)
+static int
+gml_is_srid_planar(int32_t srid)
 {
 	char *result;
 	char query[256];
@@ -703,7 +751,7 @@ static POINTARRAY* parse_gml_pos(xmlNodePtr xnode, bool *hasz)
 	POINTARRAY *dpa;
 	char *pos, *p;
 	bool digit;
-	POINT4D pt = {0, 0, 0, 0};
+	POINT4D pt;
 
 	/* HasZ, !HasM, 1 Point */
 	dpa = ptarray_construct_empty(1, 0, 1);
@@ -769,7 +817,7 @@ static POINTARRAY* parse_gml_poslist(xmlNodePtr xnode, bool *hasz)
 	char *poslist, *p;
 	int dim, gml_dim;
 	POINTARRAY *dpa;
-	POINT4D pt = {0, 0, 0, 0};
+	POINT4D pt;
 	bool digit;
 
 	/* Retrieve gml:srsDimension attribute if any */
@@ -811,7 +859,6 @@ static POINTARRAY* parse_gml_poslist(xmlNodePtr xnode, bool *hasz)
 			if (gml_dim == dim)
 			{
 				ptarray_append_point(dpa, &pt, LW_FALSE);
-				pt.x = pt.y = pt.z = pt.m = 0.0;
 				gml_dim = 0;
 			}
 			else if (*(poslist+1) == '\0')
@@ -1001,7 +1048,7 @@ static LWGEOM* parse_gml_line(xmlNodePtr xnode, bool *hasz, int *root_srid)
 static LWGEOM* parse_gml_curve(xmlNodePtr xnode, bool *hasz, int *root_srid)
 {
 	xmlNodePtr xa;
-	size_t lss;
+	int lss, last, i;
 	bool found=false;
 	gmlSrs srs;
 	LWGEOM *geom=NULL;
@@ -1057,46 +1104,37 @@ static LWGEOM* parse_gml_curve(xmlNodePtr xnode, bool *hasz, int *root_srid)
 	/* Most common case, a single segment */
 	if (lss == 1) pa = ppa[0];
 
+	/*
+	 * "The curve segments are connected to one another, with the end point
+	 *  of each segment except the last being the start point of the next
+	 *  segment"  from  ISO 19107:2003 -> 6.3.16.1 (p43)
+	 *
+	 * So we must aggregate all the segments into a single one and avoid
+	 * to copy the redundant points
+	 */
 	if (lss > 1)
 	{
-		/*
-		 * "The curve segments are connected to one another, with the end point
-		 *  of each segment except the last being the start point of the next
-		 *  segment"  from  ISO 19107:2003 -> 6.3.16.1 (p43)
-		 *
-		 * So we must aggregate all the segments into a single one and avoid
-		 * to copy the redundant points
-		 */
-		size_t i;
-		size_t cp_point_size = sizeof(POINT3D); /* All internals are done with 3D */
-		size_t final_point_size = *hasz ? sizeof(POINT3D) : sizeof(POINT2D);
-		pa = ptarray_construct(1, 0, npoints - lss + 1);
-
-		/* Copy the first linestring fully */
-		memcpy(getPoint_internal(pa, 0), getPoint_internal(ppa[0], 0), cp_point_size * (ppa[0]->npoints));
-		npoints = ppa[0]->npoints;
-		lwfree(ppa[0]);
-
-		/* For the rest of linestrings, ensure the first point matches the
-		 * last point of the previous one, and copy all points except the
-		 * first one (since it'd be repeated)
-		 */
-		for (i = 1; i < lss; i++)
+		pa = ptarray_construct(1, 0, npoints - (lss - 1));
+		for (last = npoints = i = 0; i < lss ; i++)
 		{
-			if (memcmp(getPoint_internal(pa, npoints - 1), getPoint_internal(ppa[i], 0), final_point_size))
+			if (i + 1 == lss) last = 1;
+			/* Check if segments are not disjoints */
+			if (i > 0 && memcmp( getPoint_internal(pa, npoints),
+			                     getPoint_internal(ppa[i], 0),
+			                     *hasz ? sizeof(POINT3D) : sizeof(POINT2D)))
 				gml_lwpgerror("invalid GML representation", 41);
 
-			memcpy(getPoint_internal(pa, npoints),
-			       getPoint_internal(ppa[i], 1),
-			       cp_point_size * (ppa[i]->npoints - 1));
+			/* Aggregate stuff */
+			memcpy(	getPoint_internal(pa, npoints),
+			        getPoint_internal(ppa[i], 0),
+			        ptarray_point_size(ppa[i]) * (ppa[i]->npoints + last));
 
 			npoints += ppa[i]->npoints - 1;
 			lwfree(ppa[i]);
 		}
+		lwfree(ppa);
 	}
 
-	lwfree(ppa);
-
 	parse_gml_srs(xnode, &srs);
 	if (srs.reverse_axis) pa = ptarray_flip_coordinates(pa);
 	if (srs.srid != *root_srid && *root_srid != SRID_UNKNOWN)
@@ -1807,23 +1845,33 @@ static LWGEOM* parse_gml_coll(xmlNodePtr xnode, bool *hasz, int *root_srid)
 /**
  * Read GML
  */
-static LWGEOM* lwgeom_from_gml(const char* xml)
+static LWGEOM *
+lwgeom_from_gml(const char *xml, int xml_size)
 {
 	xmlDocPtr xmldoc;
 	xmlNodePtr xmlroot=NULL;
-	int xml_size = strlen(xml);
-	LWGEOM *lwgeom;
+	LWGEOM *lwgeom = NULL;
 	bool hasz=true;
 	int root_srid=SRID_UNKNOWN;
 
 	/* Begin to Parse XML doc */
 	xmlInitParser();
-        xmldoc = xmlReadMemory(xml, xml_size, NULL, NULL, XML_PARSE_SAX1);
-	if (!xmldoc || (xmlroot = xmlDocGetRootElement(xmldoc)) == NULL)
+
+	xmldoc = xmlReadMemory(xml, xml_size, NULL, NULL, XML_PARSE_SAX1);
+	if (!xmldoc)
+	{
+		xmlCleanupParser();
+		gml_lwpgerror("invalid GML representation", 1);
+		return NULL;
+	}
+
+	xmlroot = xmlDocGetRootElement(xmldoc);
+	if (!xmlroot)
 	{
 		xmlFreeDoc(xmldoc);
 		xmlCleanupParser();
 		gml_lwpgerror("invalid GML representation", 1);
+		return NULL;
 	}
 
 	lwgeom = parse_gml(xmlroot, &hasz, &root_srid);
@@ -1832,10 +1880,12 @@ static LWGEOM* lwgeom_from_gml(const char* xml)
 	xmlCleanupParser();
 	/* shouldn't we be releasing xmldoc too here ? */
 
-
 	if ( root_srid != SRID_UNKNOWN )
 		lwgeom->srid = root_srid;
 
+	/* Should we really do this here ? */
+	lwgeom_add_bbox(lwgeom);
+
 	/* GML geometries could be either 2 or 3D and can be nested mixed.
 	 * Missing Z dimension is even tolerated inside some GML coords
 	 *
