@@ -22,20 +22,26 @@
  *
  **********************************************************************/
 
+#include <string.h>
+#include <float.h>
+#include <math.h>
+
 #include "mvt.h"
 #include "lwgeom_geos.h"
+#include "pgsql_compat.h"
 
 #ifdef HAVE_LIBPROTOBUF
-
-#if POSTGIS_PGSQL_VERSION >= 94
 #include "utils/jsonb.h"
-#endif
 
 #if POSTGIS_PGSQL_VERSION < 110
 /* See trac ticket #3867 */
 # define DatumGetJsonbP DatumGetJsonb
 #endif
 
+#define uthash_fatal(msg) lwerror("uthash: fatal error (out of memory)")
+#define uthash_malloc(sz) palloc(sz)
+#define uthash_free(ptr,sz) pfree(ptr)
+/* Note: set UTHASH_FUNCTION (not HASH_FUNCTION) to change the hash function */
 #include "uthash.h"
 
 #define FEATURES_CAPACITY_INITIAL 50
@@ -333,22 +339,16 @@ static void parse_column_keys(mvt_agg_context *ctx)
 
 	for (i = 0; i < natts; i++)
 	{
-#if POSTGIS_PGSQL_VERSION < 110
-		Oid typoid = getBaseType(ctx->column_cache.tupdesc->attrs[i]->atttypid);
-		char *tkey = ctx->column_cache.tupdesc->attrs[i]->attname.data;
-#else
-		Oid typoid = getBaseType(ctx->column_cache.tupdesc->attrs[i].atttypid);
-		char *tkey = ctx->column_cache.tupdesc->attrs[i].attname.data;
-#endif
+		Oid typoid = getBaseType(TupleDescAttr(ctx->column_cache.tupdesc, i)->atttypid);
+		char *tkey = TupleDescAttr(ctx->column_cache.tupdesc, i)->attname.data;
 
 		ctx->column_cache.column_oid[i] = typoid;
-#if POSTGIS_PGSQL_VERSION >= 94
+
 		if (typoid == JSONBOID)
 		{
 			ctx->column_cache.column_keys_index[i] = UINT32_MAX;
 			continue;
 		}
-#endif
 
 		if (ctx->geom_name == NULL)
 		{
@@ -369,11 +369,24 @@ static void parse_column_keys(mvt_agg_context *ctx)
 			}
 		}
 
-		ctx->column_cache.column_keys_index[i] = add_key(ctx, pstrdup(tkey));
+		if (ctx->id_name &&
+			(ctx->id_index == UINT32_MAX) &&
+			(strcmp(tkey, ctx->id_name) == 0) &&
+			(typoid == INT2OID || typoid == INT4OID || typoid == INT8OID))
+		{
+			ctx->id_index = i;
+		}
+		else
+		{
+			ctx->column_cache.column_keys_index[i] = add_key(ctx, pstrdup(tkey));
+		}
 	}
 
 	if (!geom_found)
 		elog(ERROR, "parse_column_keys: no geometry column found");
+
+	if (ctx->id_name != NULL && ctx->id_index == UINT32_MAX)
+		elog(ERROR, "mvt_agg_transfn: Could not find column '%s' of integer type", ctx->id_name);
 }
 
 static void encode_keys(mvt_agg_context *ctx)
@@ -547,7 +560,6 @@ static void parse_datum_as_string(mvt_agg_context *ctx, Oid typoid,
 	add_value_as_string(ctx, value, tags, k);
 }
 
-#if POSTGIS_PGSQL_VERSION >= 94
 static uint32_t *parse_jsonb(mvt_agg_context *ctx, Jsonb *jb,
 	uint32_t *tags)
 {
@@ -609,7 +621,8 @@ static uint32_t *parse_jsonb(mvt_agg_context *ctx, Jsonb *jb,
 					PointerGetDatum(v.val.numeric)));
 				d = strtod(str, NULL);
 				l = strtol(str, NULL, 10);
-				if (FP_NEQUALS(d, (double)l))
+
+				if (fabs(d - (double)l) > FLT_EPSILON)
 				{
 					MVT_PARSE_VALUE(d, mvt_kv_double_value, double_values_hash,
 						double_value, sizeof(double));
@@ -625,7 +638,45 @@ static uint32_t *parse_jsonb(mvt_agg_context *ctx, Jsonb *jb,
 
 	return tags;
 }
-#endif
+
+/**
+ * Sets the feature id. Ignores Nulls and negative values
+ */
+static void set_feature_id(mvt_agg_context *ctx, Datum datum, bool isNull)
+{
+	Oid typoid = ctx->column_cache.column_oid[ctx->id_index];
+	int64_t value = INT64_MIN;
+
+	if (isNull)
+	{
+		POSTGIS_DEBUG(3, "set_feature_id: Ignored null value");
+		return;
+	}
+
+	switch (typoid)
+	{
+	case INT2OID:
+		value = DatumGetInt16(datum);
+		break;
+	case INT4OID:
+		value = DatumGetInt32(datum);
+		break;
+	case INT8OID:
+		value = DatumGetInt64(datum);
+		break;
+	default:
+		elog(ERROR, "set_feature_id: Feature id type does not match");
+	}
+
+	if (value < 0)
+	{
+		POSTGIS_DEBUG(3, "set_feature_id: Ignored negative value");
+		return;
+	}
+
+	ctx->feature->has_id = true;
+	ctx->feature->id = (uint64_t) value;
+}
 
 static void parse_values(mvt_agg_context *ctx)
 {
@@ -661,21 +712,22 @@ static void parse_values(mvt_agg_context *ctx)
 		if (i == ctx->geom_index)
 			continue;
 
+		if (i == ctx->id_index)
+		{
+			set_feature_id(ctx, datum, cc.nulls[i]);
+			continue;
+		}
+
 		if (cc.nulls[i])
 		{
 			POSTGIS_DEBUG(3, "parse_values isnull detected");
 			continue;
 		}
 
-#if POSTGIS_PGSQL_VERSION < 110
-		key = cc.tupdesc->attrs[i]->attname.data;
-#else
-		key = cc.tupdesc->attrs[i].attname.data;
-#endif
+		key = TupleDescAttr(cc.tupdesc, i)->attname.data;
 		k = cc.column_keys_index[i];
 		typoid = cc.column_oid[i];
 
-#if POSTGIS_PGSQL_VERSION >= 94
 		if (k == UINT32_MAX && typoid != JSONBOID)
 			elog(ERROR, "parse_values: unexpectedly could not find parsed key name '%s'", key);
 		if (typoid == JSONBOID)
@@ -683,10 +735,6 @@ static void parse_values(mvt_agg_context *ctx)
 			tags = parse_jsonb(ctx, DatumGetJsonbP(datum), tags);
 			continue;
 		}
-#else
-		if (k == UINT32_MAX)
-			elog(ERROR, "parse_values: unexpectedly could not find parsed key name '%s'", key);
-#endif
 
 		switch (typoid)
 		{
@@ -718,6 +766,7 @@ static void parse_values(mvt_agg_context *ctx)
 			parse_datum_as_string(ctx, typoid, datum, tags, k);
 			break;
 		}
+
 		ctx->row_columns++;
 	}
 
@@ -739,11 +788,14 @@ lwgeom_get_basic_type(LWGEOM *geom)
 	case LINETYPE:
 	case POLYGONTYPE:
 		return geom->type;
+	case TRIANGLETYPE:
+		return POLYGONTYPE;
 	case MULTIPOINTTYPE:
 	case MULTILINETYPE:
 	case MULTIPOLYGONTYPE:
 		return geom->type - 3; /* Based on LWTYPE positions */
 	case COLLECTIONTYPE:
+	case TINTYPE:
 	{
 		uint32_t i;
 		uint8 type = 0;
@@ -998,7 +1050,12 @@ mvt_grid_and_validate_geos(LWGEOM *ng, uint8_t basic_type)
 
 		while (!valid && iterations < max_iterations)
 		{
+#if POSTGIS_GEOS_VERSION < 38
 			GEOSGeometry *geo_valid = LWGEOM_GEOS_makeValid(geo);
+#else
+			GEOSGeometry *geo_valid = GEOSMakeValid(geo);
+#endif
+
 			GEOSGeom_destroy(geo);
 			if (!geo_valid)
 				return NULL;
@@ -1066,6 +1123,48 @@ mvt_clip_and_validate_geos(LWGEOM *lwgeom, uint8_t basic_type, uint32_t extent,
 	return ng;
 }
 
+#ifdef HAVE_WAGYU
+
+#include "lwgeom_wagyu.h"
+
+static LWGEOM *
+mvt_clip_and_validate(LWGEOM *lwgeom, uint8_t basic_type, uint32_t extent, uint32_t buffer, bool clip_geom)
+{
+	GBOX clip_box = {0};
+	LWGEOM *clipped_lwgeom;
+
+	/* Wagyu only supports polygons. Default to geos for other types */
+	lwgeom = lwgeom_to_basic_type(lwgeom, POLYGONTYPE);
+	if (lwgeom->type != POLYGONTYPE && lwgeom->type != MULTIPOLYGONTYPE)
+	{
+		return mvt_clip_and_validate_geos(lwgeom, basic_type, extent, buffer, clip_geom);
+	}
+
+	if (!clip_geom)
+	{
+		/* With clipping disabled, we request a clip with the geometry bbox to force validation */
+		lwgeom_calculate_gbox(lwgeom, &clip_box);
+	}
+	else
+	{
+		clip_box.xmax = clip_box.ymax = (double)extent + (double)buffer;
+		clip_box.xmin = clip_box.ymin = -(double)buffer;
+	}
+
+	clipped_lwgeom = lwgeom_wagyu_clip_by_box(lwgeom, &clip_box);
+
+	return clipped_lwgeom;
+}
+
+#else /* ! HAVE_WAGYU */
+
+static LWGEOM *
+mvt_clip_and_validate(LWGEOM *lwgeom, uint8_t basic_type, uint32_t extent, uint32_t buffer, bool clip_geom)
+{
+	return mvt_clip_and_validate_geos(lwgeom, basic_type, extent, buffer, clip_geom);
+}
+#endif
+
 /**
  * Transform a geometry into vector tile coordinate space.
  *
@@ -1093,12 +1192,6 @@ LWGEOM *mvt_geom(LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32_t buf
 	if (lwgeom_is_empty(lwgeom))
 		return NULL;
 
-	if (width == 0 || height == 0)
-		elog(ERROR, "mvt_geom: bounds width or height cannot be 0");
-
-	if (extent == 0)
-		elog(ERROR, "mvt_geom: extent cannot be 0");
-
 	resx = width / extent;
 	resy = height / extent;
 	res = (resx < resy ? resx : resy)/2;
@@ -1121,14 +1214,14 @@ LWGEOM *mvt_geom(LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32_t buf
 	affine.yoff = -gbox->ymax * fy;
 	lwgeom_affine(lwgeom, &affine);
 
-	/* snap to integer precision, removing duplicate points */
+	/* Snap to integer precision, removing duplicate points */
 	lwgeom_grid_in_place(lwgeom, &grid);
 
-	if (lwgeom == NULL || lwgeom_is_empty(lwgeom))
+	if (!lwgeom || lwgeom_is_empty(lwgeom))
 		return NULL;
 
-	lwgeom = mvt_clip_and_validate_geos(lwgeom, basic_type, extent, buffer, clip_geom);
-	if (lwgeom == NULL || lwgeom_is_empty(lwgeom))
+	lwgeom = mvt_clip_and_validate(lwgeom, basic_type, extent, buffer, clip_geom);
+	if (!lwgeom || lwgeom_is_empty(lwgeom))
 		return NULL;
 
 	return lwgeom;
@@ -1157,6 +1250,7 @@ void mvt_agg_init_context(mvt_agg_context *ctx)
 	ctx->bool_values_hash = NULL;
 	ctx->values_hash_i = 0;
 	ctx->keys_hash_i = 0;
+	ctx->id_index = UINT32_MAX;
 	ctx->geom_index = UINT32_MAX;
 
 	memset(&ctx->column_cache, 0, sizeof(ctx->column_cache));
@@ -1300,7 +1394,7 @@ mvt_agg_context * mvt_ctx_deserialize(const bytea *ba)
 		NULL
 	};
 
-	size_t len = VARSIZE(ba) - VARHDRSZ;
+	size_t len = VARSIZE_ANY_EXHDR(ba);
 	VectorTile__Tile *tile = vector_tile__tile__unpack(&allocator, len, (uint8_t*)VARDATA(ba));
 	mvt_agg_context *ctx = palloc(sizeof(mvt_agg_context));
 	memset(ctx, 0, sizeof(mvt_agg_context));
