@@ -450,63 +450,35 @@ usage() {
 	));
 }
 
-static void calc_tile_size(
-	int dimX, int dimY,
-	int *tileX, int *tileY
-) {
-	int i = 0;
-	int j = 0;
-	int min = 30;
-	int max = 100;
+static void
+calc_tile_size(uint32_t dimX, uint32_t dimY, int *tileX, int *tileY)
+{
+	uint32_t min_tile_size = 30;
+	uint32_t max_tile_size = 300;
 
-	int d = 0;
-	double r = 0;
-	/*int _d = 0;*/
-	double _r = -1;
-	int _i = 0;
+	for (uint8_t current_dimension = 0; current_dimension <= 1; current_dimension++)
+	{
+		uint32_t img_size = (current_dimension == 0) ? dimX : dimY;
+		uint32_t best_gap = max_tile_size;
+		uint32_t best_size = img_size;
 
-	/* j = 0, X */
-	for (j = 0; j < 2; j++) {
-		_i = 0;
-		/*_d = 0;*/
-		_r = -1;
-
-		if (j < 1 && dimX <= max) {
-			*tileX = dimX;
-			continue;
-		}
-		else if (dimY <= max) {
-			*tileY = dimY;
-			continue;
-		}
-
-		for (i = max; i >= min; i--) {
-			if (j < 1) {
-				d = dimX / i;
-				r = (double) dimX / (double) i;
-
-			}
-			else {
-				d = dimY / i;
-				r = (double) dimY / (double) i;
-			}
-			r = r - (double) d;
-
-			if (
-				FLT_EQ(_r, -1) ||
-				(r < _r) ||
-				FLT_EQ(r, _r)
-			) {
-				/*_d = d;*/
-				_r = r;
-				_i = i;
+		if (img_size > max_tile_size)
+		{
+			for (uint32_t tile_size = max_tile_size; tile_size >= min_tile_size; tile_size--)
+			{
+				uint32_t gap = img_size % tile_size;
+				if (gap < best_gap)
+				{
+					best_gap = gap;
+					best_size = tile_size;
+				}
 			}
 		}
 
-		if (j < 1)
-			*tileX = _i;
+		if (current_dimension == 0)
+			*tileX = best_size;
 		else
-			*tileY = _i;
+			*tileY = best_size;
 	}
 }
 
@@ -1572,6 +1544,8 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 	int _tile_size[2] = {0, 0};
 	int xtile = 0;
 	int ytile = 0;
+	int naturalx = 1;
+	int naturaly = 1;
 	double gt[6] = {0.};
 	const char* pszProjectionRef = NULL;
 	int tilesize = 0;
@@ -1711,37 +1685,7 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 	info->dim[0] = GDALGetRasterXSize(hdsSrc);
 	info->dim[1] = GDALGetRasterYSize(hdsSrc);
 
-	/* tile size is "auto" */
-	if (
-		config->tile_size[0] == -1 &&
-		config->tile_size[1] == -1
-	) {
-		calc_tile_size(
-			info->dim[0], info->dim[1],
-			&(config->tile_size[0]), &(config->tile_size[1])
-		);
-
-		rtinfo(_("Using computed tile size: %dx%d"), config->tile_size[0], config->tile_size[1]);
-	}
-
-	/* decide on tile size */
-	if (!config->tile_size[0])
-		info->tile_size[0] = info->dim[0];
-	else
-		info->tile_size[0] = config->tile_size[0];
-	if (!config->tile_size[1])
-		info->tile_size[1] = info->dim[1];
-	else
-		info->tile_size[1] = config->tile_size[1];
-
-	/* number of tiles */
-	if ((uint32_t)info->tile_size[0] != info->dim[0])
-		ntiles[0] = (info->dim[0] + info->tile_size[0]  - 1) / info->tile_size[0];
-	if ((uint32_t)info->tile_size[1] != info->dim[1])
-		ntiles[1] = (info->dim[1] + info->tile_size[1]  - 1) / info->tile_size[1];
-
-	/* estimate size of 1 tile */
-	tilesize = info->tile_size[0] * info->tile_size[1];
+	tilesize = 0;
 
 	/* go through bands for attributes */
 	for (i = 0; i < info->nband_count; i++) {
@@ -1756,6 +1700,7 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 			GDALClose(hdsSrc);
 			return 0;
 		}
+		GDALGetBlockSize(hbandSrc, &naturalx, &naturaly);
 
 		/* convert data type to that of postgis raster */
 		info->bandtype[i] = rt_util_gdal_datatype_to_pixtype(info->gdalbandtype[i]);
@@ -1772,10 +1717,40 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 				info->nodataval[i] = 0;
 		}
 
-		/* update estimated size of 1 tile */
-		tilesize *= rt_pixtype_size(info->bandtype[i]);
+		/* update estimated byte size of 1 pixel */
+		tilesize += rt_pixtype_size(info->bandtype[i]);
 	}
 
+	/* tile size is "auto" */
+	if (config->tile_size[0] == -1 && config->tile_size[1] == -1)
+	{
+		calc_tile_size((naturalx > 1) ? (uint32_t)naturalx : info->dim[0],
+			       (naturaly > 1) ? (uint32_t)naturaly : info->dim[1],
+			       &(config->tile_size[0]),
+			       &(config->tile_size[1]));
+
+		rtinfo(_("Using computed tile size: %dx%d"), config->tile_size[0], config->tile_size[1]);
+	}
+
+	/* decide on tile size */
+	if (!config->tile_size[0])
+		info->tile_size[0] = info->dim[0];
+	else
+		info->tile_size[0] = config->tile_size[0];
+	if (!config->tile_size[1])
+		info->tile_size[1] = info->dim[1];
+	else
+		info->tile_size[1] = config->tile_size[1];
+
+	/* number of tiles */
+	if ((uint32_t)info->tile_size[0] != info->dim[0])
+		ntiles[0] = (info->dim[0] + info->tile_size[0] - 1) / info->tile_size[0];
+	if ((uint32_t)info->tile_size[1] != info->dim[1])
+		ntiles[1] = (info->dim[1] + info->tile_size[1] - 1) / info->tile_size[1];
+
+	/* estimate size of 1 tile */
+	tilesize *= info->tile_size[0] * info->tile_size[1];
+
 	/* roughly estimate size of one tile and all bands */
 	tilesize *= 1.1;
 	if (tilesize > MAXTILESIZE)
@@ -1794,6 +1769,7 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 				_tile_size[1] = info->tile_size[1];
 
 			for (xtile = 0; xtile < ntiles[0]; xtile++) {
+				int tile_is_nodata = !config->skip_nodataval_check;
 
 				/* edge x tile */
 				if (!config->pad_tile && ntiles[0] > 1 && (xtile + 1) == ntiles[0])
@@ -1844,20 +1820,23 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 
 					/* inspect each band of raster where band is NODATA */
 					if (!config->skip_nodataval_check)
-						rt_band_check_is_nodata(band);
+						tile_is_nodata = tile_is_nodata && rt_band_check_is_nodata(band);
 				}
 
 				/* convert rt_raster to hexwkb */
-				hex = rt_raster_to_hexwkb(rast, FALSE, &hexlen);
+				if (!tile_is_nodata)
+					hex = rt_raster_to_hexwkb(rast, FALSE, &hexlen);
 				raster_destroy(rast);
 
-				if (hex == NULL) {
+				if (!hex && !tile_is_nodata)
+				{
 					rterror(_("convert_raster: Could not convert PostGIS raster to hex WKB"));
 					return 0;
 				}
 
 				/* add hexwkb to tileset */
-				append_stringbuffer(tileset, hex);
+				if (!tile_is_nodata)
+					append_stringbuffer(tileset, hex);
 
 				/* flush if tileset gets too big */
 				if (tileset->length > 10) {
@@ -1891,6 +1870,7 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 				_tile_size[1] = info->tile_size[1];
 
 			for (xtile = 0; xtile < ntiles[0]; xtile++) {
+				int tile_is_nodata = !config->skip_nodataval_check;
 				/*
 				char fn[100];
 				sprintf(fn, "/tmp/tile%d.vrt", (ytile * ntiles[0]) + xtile);
@@ -1960,21 +1940,24 @@ convert_raster(int idx, RTLOADERCFG *config, RASTERINFO *info, STRINGBUFFER *til
 				for (i = 0; i < numbands; i++) {
 					band = rt_raster_get_band(rast, i);
 					if (band != NULL && !config->skip_nodataval_check)
-						rt_band_check_is_nodata(band);
+						tile_is_nodata = tile_is_nodata && rt_band_check_is_nodata(band);
 				}
 
 				/* convert rt_raster to hexwkb */
-				hex = rt_raster_to_hexwkb(rast, FALSE, &hexlen);
+				if (!tile_is_nodata)
+					hex = rt_raster_to_hexwkb(rast, FALSE, &hexlen);
 				raster_destroy(rast);
 
-				if (hex == NULL) {
+				if (!hex && !tile_is_nodata)
+				{
 					rterror(_("convert_raster: Could not convert PostGIS raster to hex WKB"));
 					GDALClose(hdsDst);
 					return 0;
 				}
 
 				/* add hexwkb to tileset */
-				append_stringbuffer(tileset, hex);
+				if (!tile_is_nodata)
+					append_stringbuffer(tileset, hex);
 
 				GDALClose(hdsDst);
 
