@@ -70,6 +70,7 @@ my $OPT_UPGRADE_PATH = '';
 our $OPT_UPGRADE_FROM = '';
 my $OPT_UPGRADE_TO = '';
 my $VERBOSE = 0;
+my $OPT_SCHEMA = 'public';
 
 GetOptions (
 	'verbose' => \$VERBOSE,
@@ -83,7 +84,8 @@ GetOptions (
 	'raster' => \$OPT_WITH_RASTER,
 	'sfcgal' => \$OPT_WITH_SFCGAL,
 	'expect' => \$OPT_EXPECT,
-	'extensions' => \$OPT_EXTENSIONS
+	'extensions' => \$OPT_EXTENSIONS,
+	'schema=s' => \$OPT_SCHEMA
 	);
 
 if ( @ARGV < 1 )
@@ -106,13 +108,18 @@ if ( $OPT_UPGRADE_PATH )
   print "Upgrade path: ${OPT_UPGRADE_FROM} --> ${OPT_UPGRADE_TO}\n";
 }
 
-if ( $OPT_EXTENSIONS )
+# Split-raster extension was introduced in PostGIS-3.0.0
+sub has_split_raster_ext
 {
-	$OPT_WITH_RASTER = 1; # implied
+  my $fullver = shift;
+  # unpackaged is always current, so does have
+  # split raster already.
+  return 1 if $fullver =~ /^unpackaged/;
+  my @ver = split(/\./, $fullver);
+  return 0 if ( $ver[0] < 3 );
+  return 1;
 }
 
-
-
 ##################################################################
 # Set the locale to "C" so error messages match
 # Save original locale to set back
@@ -147,21 +154,6 @@ my $OBJ_COUNT_POST = 0;
 
 print "PATH is $PATH\n";
 
-#foreach my $exec ( ("psql", "createdb", "createlang", "dropdb") )
-#{
-#	my $execdir = which( $exec );
-#	print "Checking for $exec ... ";
-#	if ( $execdir )
-#	{
-#		print "found $execdir\n";
-#	}
-#	else
-#	{
-#		print "failed\n";
-#		die "Unable to find $exec executable. Please ensure it is on your PATH.\n";
-#	}
-#}
-
 foreach my $exec ( ($SHP2PGSQL, $PGSQL2SHP) )
 {
 	printf "Checking for %s ... ", basename($exec);
@@ -217,7 +209,6 @@ my $REGRESS_LOG = "${TMPDIR}/regress_log";
 # Report
 print "TMPDIR is $TMPDIR\n";
 
-
 ##################################################################
 # Prepare the database
 ##################################################################
@@ -246,13 +237,14 @@ else
 	}
 	else
 	{
-		print "Database $DB already exists.\n";
-		print "Drop it, or run with the --nocreate flag to use it.\n";
-		exit(1);
+		print STDERR "Database $DB already exists, dropping.\n";
+		`dropdb $DB`;
+		create_spatial();
 	}
 }
 
 my $libver = sql("select postgis_lib_version()");
+my $defextver = sql("select default_version from pg_available_extensions where name = 'postgis'");
 
 if ( ! $libver )
 {
@@ -262,6 +254,23 @@ if ( ! $libver )
 	exit(1);
 }
 
+sub scriptdir
+{
+	my ( $version, $systemwide ) = @_;
+	my $scriptdir;
+	if ( $systemwide or ( $version and $version ne $libver ) ) {
+		my $pgis_majmin = $version;
+		$pgis_majmin =~ s/^([1-9]*\.[0-9]*).*/\1/;
+		$scriptdir = `pg_config --sharedir`;
+		chop $scriptdir;
+		$scriptdir .= "/contrib/postgis-" . $pgis_majmin;
+	} else {
+		$scriptdir = $STAGED_SCRIPTS_DIR;
+	}
+	#print "XXX: scriptdir: $scriptdir\n";
+	return $scriptdir;
+}
+
 sub semver_lessthan
 {
   my ($a,$b) = @_;
@@ -292,11 +301,37 @@ sub create_upgrade_test_objects
     exit(1);
   }
 
+  $query = "insert into upgrade_test(g1,g2) values ";
+	$query .= "('POINT(0 0)', 'LINESTRING(0 0, 1 1)'), ";
+	$query .= "('POINT(1 0)', 'LINESTRING(0 1, 1 1)');";
+  my $ret = sql($query);
+  unless ( $ret =~ /^INSERT/ ) {
+    `dropdb $DB`;
+    print "\nSomething went wrong populating upgrade_test table: $ret.\n";
+    exit(1);
+  }
+
+  # my $query = "create view upgrade_view_test as ";
+  # $query .= "select st_union(g1) from upgrade_test;";
+  # my $ret = sql($query);
+  # unless ( $ret =~ /^CREATE/ ) {
+  #   `dropdb $DB`;
+  #   print "\nSomething went wrong creating upgrade_view_test view: $ret.\n";
+  #   exit(1);
+  # }
+
   if ( $OPT_WITH_RASTER )
   {
-    $query = "insert into upgrade_test(r) ";
-    $query .= "select ST_AddBand(ST_MakeEmptyRaster(10, 10, 1, 1, 2, 2, 0, 0,4326), 1, '8BSI'::text, -129, NULL);";
-    $query .= "set client_min_messages to error; select AddRasterConstraints('upgrade_test', 'r')";
+    $query = "UPDATE upgrade_test SET r = ";
+    $query .= " ST_AddBand(ST_MakeEmptyRaster(10, 10, 1, 1, 2, 2, 0, 0,4326), 1, '8BSI'::text, -129, NULL);";
+    $ret = sql($query);
+    unless ( $ret =~ /^UPDATE/ ) {
+      `dropdb $DB`;
+      print "\nSomething went wrong setting raster into upgrade_test table: $ret.\n";
+      exit(1);
+    }
+
+    $query = "set client_min_messages to error; select AddRasterConstraints('upgrade_test', 'r')";
     $ret = sql($query);
     unless ( $ret =~ /^t$/ ) {
       `dropdb $DB`;
@@ -321,6 +356,13 @@ sub drop_upgrade_test_objects
 {
   # TODO: allow passing the "upgrade-cleanup" script via commandline
 
+  # my $ret = sql("drop view upgrade_view_test;");
+  # unless ( $ret =~ /^DROP/ ) {
+  #   `dropdb $DB`;
+  #   print "\nSomething went wrong dropping spatial view: $ret.\n";
+  #   exit(1);
+  # }
+
   my $ret = sql("drop table upgrade_test;");
   unless ( $ret =~ /^DROP/ ) {
     `dropdb $DB`;
@@ -407,10 +449,14 @@ print "\nRunning tests\n\n";
 
 foreach $TEST (@ARGV)
 {
+	my $TEST_OBJ_COUNT_PRE;
+	my $TEST_OBJ_COUNT_POST;
+
 	# catch a common mistake (strip trailing .sql)
 	$TEST =~ s/.sql$//;
 
 	start_test($TEST);
+	$TEST_OBJ_COUNT_PRE = count_postgis_objects();
 
 	# Check for a "-pre.pl" file in case there are setup commands
     eval_file("${TEST}-pre.pl");
@@ -465,6 +511,14 @@ foreach $TEST (@ARGV)
 	# Check for a "-post.pl" file in case there are teardown commands
     eval_file("${TEST}-post.pl");
 
+	$TEST_OBJ_COUNT_POST = count_postgis_objects();
+
+	if ( $TEST_OBJ_COUNT_POST != $TEST_OBJ_COUNT_PRE )
+	{
+		fail("PostGIS object count pre-test ($TEST_OBJ_COUNT_POST) != post-test ($TEST_OBJ_COUNT_PRE)");
+		return 0;
+	}
+
 }
 
 
@@ -524,7 +578,14 @@ Options:
   -v, --verbose   be verbose about failures
   --nocreate      do not create the regression database on start
   --upgrade       source the upgrade scripts on start
-  --upgrade-path  upgrade path, format <from>--<to>
+  --upgrade-path  upgrade path, format <from>--<to>.
+                  <from> can be specified as "unpackaged<version>"
+                         to specify a script version to start from.
+                  <to> can be specified as ":auto" to request
+                       upgrades to default version, and be appended
+                       a question mark (ie: ":auto!" or "3.0.0!") to
+                       request upgrade via postgis_extensions_upgrade()
+                       if available.
   --dumprestore   dump and restore spatially-enabled db before running tests
   --nodrop        do not drop the regression database on exit
   --raster        load also raster extension
@@ -612,7 +673,7 @@ sub run_simple_sql
 
 	# Dump output to a temp file.
 	my $tmpfile = sprintf("%s/test_%s_tmp", $TMPDIR, $RUN);
-	my $cmd = "psql -v \"VERBOSITY=terse\" -tXA $DB < $sql > $tmpfile 2>&1";
+	my $cmd = "psql -v \"VERBOSITY=terse\" -tXAq $DB < $sql > $tmpfile 2>&1";
 	#print($cmd);
 	my $rv = system($cmd);
 	# Check if psql errored out.
@@ -641,7 +702,7 @@ sub run_simple_sql
 sub drop_table
 {
 	my $tblname = shift;
-	my $cmd = "psql -tXA -d $DB -c \"DROP TABLE IF EXISTS $tblname\" >> $REGRESS_LOG 2>&1";
+	my $cmd = "psql -tXAq -d $DB -c \"DROP TABLE IF EXISTS $tblname\" >> $REGRESS_LOG 2>&1";
 	my $rv = system($cmd);
 	die "Could not run: $cmd\n" if $rv;
 }
@@ -704,22 +765,13 @@ sub run_simple_test
 	mkpath($betmpdir);
 	chmod 0777, $betmpdir;
 
-	my $scriptdir;
-	if ( $OPT_EXTENSIONS ) {
-		# TODO: allow override this default with env variable ?
-		my $pgis_majmin = $libver;
-		$pgis_majmin =~ s/^([1-9]*\.[1-9]*).*/\1/;
-		$scriptdir = `pg_config --sharedir`;
-		chop $scriptdir;
-		$scriptdir .= "/contrib/postgis-" . $pgis_majmin;
-	} else {
-		$scriptdir = $STAGED_SCRIPTS_DIR;
-	}
+	my $scriptdir = scriptdir($libver, $OPT_EXTENSIONS);
 	my $cmd = "psql -v \"VERBOSITY=terse\""
           . " -v \"tmpfile='$tmpfile'\""
           . " -v \"scriptdir=$scriptdir\""
           . " -v \"regdir=$REGDIR\""
-          . " -tXA $DB < $sql > $outfile 2>&1";
+          . " -v \"schema=$OPT_SCHEMA.\""
+          . " -tXAq $DB < $sql > $outfile 2>&1";
 	my $rv = system($cmd);
 
 	# Check for ERROR lines
@@ -827,7 +879,7 @@ sub run_loader_and_check_output
 
 		if ( $rv )
 		{
-			fail(" $description: running shp2pgsql", "$errfile");
+			fail(" $description: running $cmd", "$errfile");
 			return 0;
 		}
 
@@ -1214,20 +1266,36 @@ sub count_db_objects
 		select count(*) from pg_opclass union all
 		select count(*) from pg_namespace
 			where nspname NOT LIKE 'pg_%' union all
-		select count(*) from pg_opfamily )
+		select count(*) from pg_opfamily
+		)
 		select sum(count) from counts");
 
  	return $count;
 }
 
 
+##################################################################
+# Count postgis objects
+##################################################################
+sub count_postgis_objects
+{
+	my $count = sql("WITH counts as (
+		select count(*) from spatial_ref_sys
+		)
+		select sum(count) from counts");
+
+ 	return $count;
+}
+
+
+
 ##################################################################
 # Create the spatial database
 ##################################################################
 sub create_db
 {
-	my $cmd = "createdb --encoding=UTF-8 --template=template0 --lc-collate=C $DB > $REGRESS_LOG";
-	return system($cmd);
+	my $createcmd = "createdb --encoding=UTF-8 --template=template0 --lc-collate=C $DB > $REGRESS_LOG";
+	return system($createcmd);
 }
 
 sub create_spatial
@@ -1235,10 +1303,7 @@ sub create_spatial
 	my ($cmd, $rv);
 	print "Creating database '$DB' \n";
 
-  $rv = create_db();
-
-	$cmd = "createlang plpgsql $DB >> $REGRESS_LOG 2>&1";
-	$rv = system($cmd);
+  	$rv = create_db();
 
 	# Count database objects before installing anything
 	$OBJ_COUNT_PRE = count_db_objects();
@@ -1274,7 +1339,7 @@ sub load_sql_file
 		if ( $rv )
 		{
 		  fail "Error encountered loading $file", $REGRESS_LOG;
-		  exit 1
+		  exit 1;
 		}
 	}
 	return 1;
@@ -1286,9 +1351,10 @@ sub prepare_spatial_extensions
 	# ON_ERROR_STOP is used by psql to return non-0 on an error
 	my $psql_opts = "--no-psqlrc --variable ON_ERROR_STOP=true";
 	my $sql = "CREATE EXTENSION postgis";
+
 	if ( $OPT_UPGRADE_FROM ) {
-		if ( $OPT_UPGRADE_FROM eq "unpackaged" ) {
-			prepare_spatial();
+		if ( $OPT_UPGRADE_FROM =~ /^unpackaged(.*)/ ) {
+			prepare_spatial($1);
 			return;
 		}
 		$sql .= " VERSION '" . $OPT_UPGRADE_FROM . "'";
@@ -1321,9 +1387,31 @@ sub prepare_spatial_extensions
 		}
  	}
 
+	if ( $OPT_WITH_RASTER && (
+		# NOTE: this code is assuming that the default version
+		# (!$OPT_UPGRADE_FROM) has split raster extension
+		! $OPT_UPGRADE_FROM ||
+		has_split_raster_ext($OPT_UPGRADE_FROM)
+	) )
+	{
+		my $sql = "CREATE EXTENSION postgis_raster";
+		if ( $OPT_UPGRADE_FROM ) {
+			$sql .= " VERSION '" . $OPT_UPGRADE_FROM . "'";
+		}
+
+		print "Preparing db '${DB}' using: ${sql}\n";
+
+ 		$cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
+		$rv = system($cmd);
+		if ( $rv ) {
+			fail "Error encountered creating EXTENSION POSTGIS_RASTER", $REGRESS_LOG;
+			die;
+		}
+ 	}
+
 	if ( $OPT_WITH_SFCGAL )
 	{
-		do {
+		{
 			my $sql = "CREATE EXTENSION postgis_sfcgal";
 			if ( $OPT_UPGRADE_FROM ) {
 				if ( semver_lessthan($OPT_UPGRADE_FROM, "2.2.0") )
@@ -1343,7 +1431,7 @@ sub prepare_spatial_extensions
 				fail "Error encountered creating EXTENSION POSTGIS_SFCGAL", $REGRESS_LOG;
 				die;
 			}
-		} while (0);
+		}
 	}
 
  	return 1;
@@ -1352,33 +1440,38 @@ sub prepare_spatial_extensions
 # Prepare the database for spatial operations (old method)
 sub prepare_spatial
 {
+	my $version = shift;
+	my $scriptdir = scriptdir($version);
+	print "Loading unpackaged components from $scriptdir\n";
+
 	print "Loading PostGIS into '${DB}' \n";
 
 	# Load postgis.sql into the database
-	load_sql_file("${STAGED_SCRIPTS_DIR}/postgis.sql", 1);
-	load_sql_file("${STAGED_SCRIPTS_DIR}/postgis_comments.sql", 0);
-	load_sql_file("${STAGED_SCRIPTS_DIR}/postgis_proc_set_search_path.sql", 0);
+	load_sql_file("${scriptdir}/postgis.sql", 1);
+	load_sql_file("${scriptdir}/postgis_comments.sql", 0);
+	load_sql_file("${scriptdir}/postgis_proc_set_search_path.sql", 0);
+	load_sql_file("${scriptdir}/spatial_ref_sys.sql", 0);
 
 	if ( $OPT_WITH_TOPO )
 	{
 		print "Loading Topology into '${DB}'\n";
-		load_sql_file("${STAGED_SCRIPTS_DIR}/topology.sql", 1);
-		load_sql_file("${STAGED_SCRIPTS_DIR}/topology_comments.sql", 0);
+		load_sql_file("${scriptdir}/topology.sql", 1);
+		load_sql_file("${scriptdir}/topology_comments.sql", 0);
 	}
 
 	if ( $OPT_WITH_RASTER )
 	{
 		print "Loading Raster into '${DB}'\n";
-		load_sql_file("${STAGED_SCRIPTS_DIR}/rtpostgis.sql", 1);
-		load_sql_file("${STAGED_SCRIPTS_DIR}/raster_comments.sql", 0);
-		load_sql_file("${STAGED_SCRIPTS_DIR}/rtpostgis_proc_set_search_path.sql", 0);
+		load_sql_file("${scriptdir}/rtpostgis.sql", 1);
+		load_sql_file("${scriptdir}/raster_comments.sql", 0);
+		load_sql_file("${scriptdir}/rtpostgis_proc_set_search_path.sql", 0);
 	}
 
 	if ( $OPT_WITH_SFCGAL )
 	{
 		print "Loading SFCGAL into '${DB}'\n";
-		load_sql_file("${STAGED_SCRIPTS_DIR}/sfcgal.sql", 1);
-		load_sql_file("${STAGED_SCRIPTS_DIR}/sfcgal_comments.sql", 0);
+		load_sql_file("${scriptdir}/sfcgal.sql", 1);
+		load_sql_file("${scriptdir}/sfcgal_comments.sql", 0);
 	}
 
 	return 1;
@@ -1455,12 +1548,45 @@ sub upgrade_spatial_extensions
 {
     # ON_ERROR_STOP is used by psql to return non-0 on an error
     my $psql_opts = "--no-psqlrc --variable ON_ERROR_STOP=true";
-    my $nextver = $OPT_UPGRADE_TO ? "${OPT_UPGRADE_TO}" : "${libver}next";
-    my $sql = "ALTER EXTENSION postgis UPDATE TO '${nextver}'";
+    my $sql;
+    my $upgrade_via_function = 0;
 
-    if ( $OPT_UPGRADE_FROM eq "unpackaged" ) {
+    if ( $OPT_UPGRADE_TO =~ /!$/ )
+    {
+      $OPT_UPGRADE_TO =~ s/!$//;
+      my $from = $OPT_UPGRADE_FROM;
+      $from =~ s/^unpackaged//;
+      if ( ! $from || ! semver_lessthan($from, "3.0.0") )
+      {
+        $upgrade_via_function = 1;
+      }
+      else
+      {
+        print "WARNING: postgis_extensions_upgrade()".
+              " not available or functional in version $from.".
+              " We'll use manual upgrade.\n";
+      }
+    }
+
+    if ( $OPT_UPGRADE_TO =~ /^:auto/ )
+    {
+      $OPT_UPGRADE_TO = $defextver;
+    }
+
+    my $nextver = $OPT_UPGRADE_TO ? "${OPT_UPGRADE_TO}" : "${libver}next";
+
+    if ( $upgrade_via_function )
+    {
+      $sql = "SELECT postgis_extensions_upgrade()";
+    }
+    elsif ( $OPT_UPGRADE_FROM =~ /^unpackaged/ )
+    {
       $sql = "CREATE EXTENSION postgis VERSION '${nextver}' FROM unpackaged";
     }
+    else
+    {
+      $sql = "ALTER EXTENSION postgis UPDATE TO '${nextver}'";
+    }
 
     print "Upgrading PostGIS in '${DB}' using: ${sql}\n" ;
 
@@ -1472,15 +1598,85 @@ sub upgrade_spatial_extensions
       die;
     }
 
+    # Handle raster split if coming from pre-split and going
+    # to splitted raster
+    if ( $OPT_UPGRADE_FROM && has_split_raster_ext($OPT_UPGRADE_TO) &&
+         ! has_split_raster_ext($OPT_UPGRADE_FROM) )
+    {
+      if ( $OPT_WITH_RASTER )
+      {
+        # upgrade of postgis must have unpackaged raster, so
+        # we create it again here
+        my $sql = "CREATE EXTENSION postgis_raster VERSION '${nextver}' FROM unpackaged";
+
+        print "Upgrading PostGIS Raster in '${DB}' using: ${sql}\n" ;
+
+        my $cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
+        my $rv = system($cmd);
+        if ( $rv ) {
+          fail "Error encountered creating EXTENSION POSTGIS_RASTER from unpackaged on upgrade", $REGRESS_LOG;
+          die;
+        }
+      }
+      else
+      {
+        # Raster support was not requested, so drop it if
+        # left unpackaged
+        print "Packaging PostGIS Raster in '${DB}' for later drop using: ${sql}\n" ;
+
+        $sql = "CREATE EXTENSION postgis_raster VERSION '${nextver}' FROM unpackaged";
+        $cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
+        $rv = system($cmd);
+        if ( $rv ) {
+          fail "Error encountered creating EXTENSION POSTGIS_RASTER from unpackaged on upgrade", $REGRESS_LOG;
+          die;
+        }
+
+        print "Dropping PostGIS Raster in '${DB}' using: ${sql}\n" ;
+
+        $sql = "DROP EXTENSION postgis_raster";
+        $cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
+        $rv = system($cmd);
+        if ( $rv ) {
+          fail "Error encountered dropping EXTENSION POSTGIS_RASTER on upgrade", $REGRESS_LOG;
+          die;
+        }
+      }
+    }
+
+    if ( $upgrade_via_function )
+    {
+      # The function does everything
+      return 1;
+    }
+
+    if ( $OPT_WITH_RASTER )
+    {
+        my $sql = "ALTER EXTENSION postgis_raster UPDATE TO '${nextver}'";
+
+        if ( $OPT_UPGRADE_FROM =~ /^unpackaged/ ) {
+          $sql = "CREATE EXTENSION postgis_raster VERSION '${nextver}' FROM unpackaged";
+        }
+
+        print "Upgrading PostGIS Raster in '${DB}' using: ${sql}\n" ;
+
+        my $cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
+        my $rv = system($cmd);
+        if ( $rv ) {
+          fail "Error encountered altering EXTENSION POSTGIS_RASTER", $REGRESS_LOG;
+          die;
+        }
+    }
+
     if ( $OPT_WITH_TOPO )
     {
       my $sql = "ALTER EXTENSION postgis_topology UPDATE TO '${nextver}'";
 
-			if ( $OPT_UPGRADE_FROM eq "unpackaged" ) {
+			if ( $OPT_UPGRADE_FROM =~ /^unpackaged/ ) {
 				$sql = "CREATE EXTENSION postgis_topology VERSION '${nextver}' FROM unpackaged";
 			}
 
-			print "Upgrading PostGIS Topology in '${DB}' using: ${sql}\n" ;
+      print "Upgrading PostGIS Topology in '${DB}' using: ${sql}\n";
 
       my $cmd = "psql $psql_opts -c \"" . $sql . "\" $DB >> $REGRESS_LOG 2>&1";
       my $rv = system($cmd);
@@ -1494,7 +1690,7 @@ sub upgrade_spatial_extensions
     {
 			my $sql;
 
-			if ( $OPT_UPGRADE_FROM eq "unpackaged" ) {
+			if ( $OPT_UPGRADE_FROM =~ /^unpackaged/ ) {
 				$sql = "CREATE EXTENSION postgis_sfcgal VERSION '${nextver}' FROM unpackaged";
 			}
 			elsif ( $OPT_UPGRADE_FROM && semver_lessthan($OPT_UPGRADE_FROM, "2.2.0") )
@@ -1566,10 +1762,19 @@ sub drop_spatial_extensions
         $ok = 0 if $rv;
     }
 
+    if ( $OPT_WITH_RASTER )
+    {
+        $cmd = "psql $psql_opts -c \"DROP EXTENSION IF EXISTS postgis_raster;\" $DB >> $REGRESS_LOG 2>&1";
+        $rv = system($cmd);
+      	$ok = 0 if $rv;
+    }
+
     $cmd = "psql $psql_opts -c \"DROP EXTENSION postgis\" $DB >> $REGRESS_LOG 2>&1";
     $rv = system($cmd);
-  	die "\nError encountered dropping EXTENSION POSTGIS, see $REGRESS_LOG for details\n\n"
-  	    if $rv;
+    if ( $rv ) {
+        fail "Error encountered dropping EXTENSION POSTGIS", $REGRESS_LOG;
+        die;
+    }
 
     return $ok;
 }
